//#define MULTISAMPLE 4
cbuffer Basic : register (b0)
{
  float4x4 world;
  float4x4 worldViewProj;
}

#define LightType_Ambient     0
#define LightType_Directional 1

#define Max_Lights 8
struct LightData
{
  int type;
  float3 positionOrDirection;
  float3 color;
};

cbuffer Light : register(b1)
{
  LightData lights[Max_Lights] : packoffset(c0);
  int numLights                : packoffset(c16);
};



struct VsOutput
{
  float4 position  : SV_Position;
  float3 normal    : NORMAL;
  float3 tangent   : TANGENT;
  float3 bitangent : BITANGENT;
  float2 tex0      : TEXCOORD0;
};

Texture2D txDiffuse : register(t0);
Texture2D txNormal : register(t1);

SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};


VsOutput VS(float4 pos : SV_Position, float3 normal : NORMAL, float3 tangent : TANGENT, float3 bitangent : BITANGENT, float2 tex0 : TEXCOORD0)
{
  VsOutput o;
  
  o.position = mul(pos, worldViewProj);
  o.normal = normal;
  o.tangent = tangent;
  o.bitangent = bitangent;
  o.tex0 = tex0;
  return o;
}

struct PSOutput
{
  float4 unlitAndAlpha            : SV_Target0; // RGB = unlit color, A = alpha
  float4 normalAndMultisampleFlag : SV_Target1; // RGB = normal XYZ,  A = multisample flag (not written by the MRT pass)
  float4 diffuseAndMaterialId     : SV_Target2; // RGB = diffuse color, A = material ID
  float4 depth                    : SV_Target3; // R = depth, GBA = nothing.
};

PSOutput PS(VsOutput i)
{
  PSOutput o;
  o.unlitAndAlpha = float4(0,0,0,txDiffuse.Sample( samLinear, i.tex0 ).a);
  
  float3 normal;
  float3 mapNormal = txNormal.Sample( samLinear, i.tex0 ).rgb*2-1; //tex2D( NormalSampler, texCoord );
  normal.x = dot(i.tangent,  mapNormal);
  normal.y = dot(i.bitangent, mapNormal);
  normal.z = dot(i.normal,   mapNormal);
  normal = normalize(normal);
  
  o.normalAndMultisampleFlag = float4(normal*0.5f+0.5f,0); //float4(i.normal*0.5f+0.5f, 0); //float4(txNormal.Sample(samLinear, i.tex0).xyz, 0);
  o.diffuseAndMaterialId = float4(txDiffuse.Sample( samLinear, i.tex0 ).rgb, 0);
  o.depth.rgba = i.position.z;
  return o;
}


struct LightVsOutput
{
  float4 position : SV_Position;
};

/*LightVsOutput LightVS(float4 pos : SV_Position)
{
  LightVsOutput o;
  o.position = float4(pos.xy,0.5, 1);
  return o;
}*/

LightVsOutput LightVS(uint id : SV_VertexID)
{
  LightVsOutput o;
  switch(id)
  {
    case 0:
      o.position = float4( 1, -1, 0.5, 1);
      break;
    case 1:
    case 5:
      o.position = float4(-1, -1, 0.5, 1);
      break;
    case 2:
    case 4:
      o.position = float4( 1,  1, 0.5, 1);
      break;
    default:
      o.position = float4(-1,  1, 0.5, 1);
      break;
  }
  return o;
}
#ifdef MULTISAMPLE
Texture2DMS<float4, MULTISAMPLE> normalAndMultisample : register(t0);
Texture2DMS<float4, MULTISAMPLE> diffuseAndMaterialId : register(t1);
Texture2DMS<float, MULTISAMPLE>  depthTexture : register(t2);
#else
Texture2D<float4> normalAndMultisample : register(t0);
Texture2D<float4> diffuseAndMaterialId : register(t1);
Texture2D<float>   depthTexture : register(t2);
#endif

#ifdef MULTISAMPLE
float4 TestPS(LightVsOutput i) : SV_Target
{
  float3 baseDiffuse = diffuseAndMaterialId.Load(int3((int2)i.position.xy, 0), 0).rgb;
  float  baseDepth   = depthTexture.Load(int3((int2)i.position.xy, 0), 0).x;
  
  float4 outer = float4(0,0,0,0);

  for(int sample = 1; sample < MULTISAMPLE; sample++)
  {
    float  depth   = depthTexture.Load(int3((int2)i.position.xy, 0), sample).x;
    float3 diffuse = diffuseAndMaterialId.Load(int3((int2)i.position.xy, 0), sample).rgb;
    if((diffuse != baseDiffuse) || (depth != baseDepth))
      outer.w = 1;
  }
  return outer;
}
#else
float4 TestPS(LightVsOutput i) : SV_Target
{
  return float4(0,0,0,0);
}
#endif

float4 LightPS(LightVsOutput i) : SV_Target
{
  float3 fullColor = float3(0,0,0);
  
  #ifdef MULTISAMPLE
  float needsMulti =  normalAndMultisample.Load(int3((int2)i.position.xy, 0), 0).w;
  
  if(needsMulti)
  {
    for(int sample = 0; sample < MULTISAMPLE; sample++)
    {
      float3 normal  = normalAndMultisample.Load(int3((int2)i.position.xy, 0), sample).xyz*2-1;
      float3 diffuse = diffuseAndMaterialId.Load(int3((int2)i.position.xy, 0), sample).rgb;
      float  depth   = depthTexture.Load(int3((int2)i.position.xy, 0), sample).x;
     
      float3 color = float3(0,0,0);
      for(int loop = 0; loop < numLights; loop++)
      {
        switch(lights[loop].type)
        {
          case LightType_Ambient:
            color += lights[loop].color * diffuse;
            break;
          case LightType_Directional:
            float lightAmount = max(0,dot(normal, lights[loop].positionOrDirection));
            color += lights[loop].color * diffuse * lightAmount;
            break;
        }
      }
      fullColor += color;
    }
    fullColor *= 1.0f / (float)MULTISAMPLE;
    //fullColor *= float3(1,0,0);
  }
  else
  #endif
  {
    float3 normal  = normalAndMultisample.Load(int3((int2)i.position.xy, 0), 0).xyz*2-1;
    float3 diffuse = diffuseAndMaterialId.Load(int3((int2)i.position.xy, 0), 0).rgb;
    float  depth   = depthTexture.Load(int3((int2)i.position.xy, 0), 0).x;
   
    for(int loop = 0; loop < numLights; loop++)
    {
      switch(lights[loop].type)
      {
        case LightType_Ambient:
          fullColor += lights[loop].color * diffuse;
          break;
        case LightType_Directional:
          float lightAmount = max(0,dot(normal, lights[loop].positionOrDirection));
          fullColor += (lights[loop].color * diffuse * lightAmount);
          break;
      }
    }
    #ifdef MULTISAMPLE
    //fullColor *= float3(0,0,1);
    #endif
  }
  
  return float4(fullColor, 1);
}

RasterizerState RasterMulti
{
  #ifdef MULTISAMPLE
    MultisampleEnable = TRUE;
  #else
    MultisampleEnable = FALSE;
  #endif
};

technique10 Geometry
{
  pass Main
  {
    SetRasterizerState(RasterMulti);
    SetVertexShader( CompileShader( vs_4_0, VS() ) );
    SetGeometryShader( NULL );
    SetPixelShader( CompileShader( ps_4_0, PS() ) );
  }
}


BlendState TestBlend
{
  RenderTargetWriteMask[0] = 0x08;
};

technique10 MultisampleTest
{
  pass Main
  {
    SetRasterizerState(RasterMulti);
    SetBlendState( TestBlend, float4(0,0,0,0), 0xFFFFFFFF );
    SetVertexShader( CompileShader( vs_4_0, LightVS() ) );
    SetGeometryShader( NULL );
    SetPixelShader( CompileShader( ps_4_0, TestPS() ) );
  }
}  


BlendState LightBlend
{
  BlendEnable[0] = TRUE;
  DestBlend = ONE;
};

technique10 LightRender
{
  pass Main
  {
    SetRasterizerState(RasterMulti);
    SetBlendState( LightBlend, float4(0,0,0,0), 0xFFFFFFFF );
    SetVertexShader( CompileShader( vs_4_0, LightVS() ) );
    SetGeometryShader( NULL );
    SetPixelShader( CompileShader( ps_4_0, LightPS() ) );
  }
}  
    